/*  Header file for lik.c                              */

/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#ifndef _LIK_H_
#define _LIK_H_

#define MAX_NAME_LEN 64
#define MAX_GLOBAL_MLE 64
#define MAX_CATEGORIES 32
#define MAX_CLOCK_RATES 5

#define EPS 1.0E-5			/* a small number - always useful */
#define BIG 1.0E5			/* a big number - always useful */
#define FTOL 1.0E-4			/* tolerance for brent/golden search */
#define FTOL2 1.0E-6		/* final tolerance for brent/golden search */
#define STOL 1.0E-4			/* tolerence for smoothing */
#define QTOL 1.0E-2			/* tolerence for quick smoothing */

typedef struct TNode TNode;
struct TNode {
	short 	tipNo, 			/* the tip (taxon) number, -1 if not tip */
			nodeNo, 		/* the internal node number, -1 if tip */
			branchNo;		/* the branch number */
				
	TNode 	*next,			/* the next node in triplet (if not tip) */
			*branch;		/* the node at the other end of the branch */
			
	double *branchLength;	/* pointer to the branch length */
	double *Likelihood; 	/* array of partial likelihoods for each pattern */

	char *sequence;
};

typedef struct TTree TTree;
struct TTree {
	short 	isRooted,		/* whether tree is rooted */
			numTips, 		/* the number of tip */
			numNodes, 		/* the number of internal nodes */
			totalNodes;		/* the total number of node structs = 3*numNodes + numTips */
			
	TNode	*nodes;			/* array of node structures (first numTips are the tips) */
	TNode	*start;			/* a starting position for calculating likelihood */
	
	TNode	**branches;		/* array of pointers to TNodes on one side of a branch */

	double	lnL;			/* the ln likelihood of the tree */
};


enum treeTypes {UnrootedTree, RootedTree};

enum nucleotides {A, C, G, T};

extern int numTaxa, numBases, numTrees, numDatasets, numSets, numPatterns, ambiguousBases;
extern short *patternIndex;
extern short *patternWeights[MAX_CATEGORIES];
extern double *patternLik[MAX_CATEGORIES];
extern short *categories;

extern char **names;
extern char **sequences;
extern char **patterns;

extern double globalMLE[MAX_GLOBAL_MLE*10];
extern int globalMask[MAX_GLOBAL_MLE*10];
extern int numGlobalMLE, numModelCats, estimateGlobals;

extern double *transMatrix[MAX_CATEGORIES];
extern double *branchValue;
extern char *branchChanged;
extern int numBranches;

extern int numCategories, gammaRates, codonRates;
extern int codonCats[3];
extern double freqRate[MAX_CATEGORIES], siteRate[MAX_CATEGORIES];
extern double alpha;


extern int verbose, verboseMemory, verboseSiteLik, quiet;
extern int debug;
extern long totalMem;
extern double freq[4], addFreq[4], freqA, freqC, freqG, freqT;

extern void (*InitFunc)(TNode *);
extern void (*GlobalFunc)(int);
extern void (*LabelFunc)(void);
extern void (*DoneFunc)(TNode *);
extern void (*MatrixFunc)(double *, int, int);
extern void (*AllMatricesFunc)(double *, int);


extern int model;

enum {
	F84,
	HKY,
	REV,
	numModels
};

/* used to look up transition type */
/* 0 - stays the same, 1 - transition and 2 - transversion */
extern short transType[16];

/* prototypes */
int FindTipName(char *aName);
void *AllocMem(long n, char *name, char *func, int showInfo);
void *CAllocMem(long n, char *name, char *func, int showInfo);
void AllocateMemory();
void FreeMemory();
void CreatePatterns();
void AllocateTreeLikelihoods(TTree *tree);
double SetBaseLik(char base, double *site, int seqNum);
void SetTipLikelihoods(TTree *tree);
void SetModel(int printHelp);
int ReadModelParam(int argc, char **argv, int *index);
int FinishModelParam();

int GetDoubleParams(int argc, char **argv, int *argn, char *pos, int numParams, double *params);
int GetIntParams(int argc, char **argv, int *argn, char *pos, int numParams, int *params);
int GetStrParam(int argc, char **argv, int *argn, char *pos, char *param, int len);

#endif /* _LIK_H_ */
