/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "random.h"

#ifdef __MWERKS__
#include <console.h>
#endif	

/*#include <Profiler.h>*/

#include "lik.h"
#include "calclik.h"
#include "anomaly.h"
#include "tree.h"
#include "fF84.h"
#include "gamma.h"

#define PROGRAM_NAME "plato"
#define VERSION_NUMBER "Version 2.0"

char *ModelNames[numModels]={
	"F84",
	"HKY",
	"REV"
};

int model, printHelp, treeFile;
char treeFileName[256];
FILE *tree_fv;

/* prototypes */

static void PrintTitle();
static void PrintUsage();
void PrintVerbose(TTree *tree);
void PrintSiteLik(FILE *fv);
void ReadParams();
void ReadFileParams();
void ReadFile();
int OpenTreeFile();
TTree *ReadNextTree(TTree *tree);
void CloseTreeFile();
void WriteFastDNAmlFile(FILE *fv, int userTree);

/* functions */

static void PrintTitle()
{
	fprintf(stderr, "Partial Likelihood Anomalies detected Through Optimization - PLATO\n");
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 1998 Nick Grassly and Andrew Rambaut\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{
	fprintf(stderr, "Usage: %s [-mMODEL] [+c|-c #1 #2 [#3]] [-g #] [-a #] [-r #] [-l #] [-h|-help] [parameters]\n          [-u[l]] [-v[b][f][m][p][s][w]|-q] [treefile]\n", PROGRAM_NAME);
	fprintf(stderr, "  -m: MODEL = F84, HKY or REV [default=F84]\n");
	fprintf(stderr, "  -c: #1 #2 #3 = codon category rates [default = equal]\n");
	fprintf(stderr, "  -g: # = num categories for gamma rate heterogeneity [default = homogeneity]\n");
	fprintf(stderr, "  -a: # = gamma shape (alpha) for rate heterogeneity [default = homogeneity]\n");
	fprintf(stderr, "  -r: # = number of replicates for the MC simulation [default=100]\n");
	fprintf(stderr, "  -l: # = minimum window size [default=5]\n");
	fprintf(stderr, "  -h,-help = Give this help message (with details of selected model)\n");
	fprintf(stderr, "  parameters = Specific to model.  To get details use: %s -mMODEL -help \n", PROGRAM_NAME);
	fprintf(stderr, "  -v: Verbose information\n");
	fprintf(stderr, "    m Memory usage information\n");
	fprintf(stderr, "    s Write likelihoods for each site\n");
	fprintf(stderr, "  -q: Quiet\n");
	fprintf(stderr, "  treefile: Optional file containing tree in PHYLIP format\n");
}


void ReadParams(int argc, char **argv)
{
	int i, j;
	char ch;
	char *P, st[4];
	double sumRates;
	
	model=F84;
	printHelp=0;
	treeFile=0;
	
	numReps=100;
	minWinSize=5;
	numTestRegions=0;
	
	verbose=0;
	verboseMemory=0;
	verboseSiteLik=0;
	quiet=0;

	debug=0;
	
	numCategories=1;
	codonCats[0]=codonCats[1]=codonCats[2]=0;
	for (i=0; i<MAX_CATEGORIES; i++)
		siteRate[i]=1.0;
	codonRates=0;
	gammaRates=0;
	alpha=1.0;
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (P[0]=='-') {
			ch=toupper(P[1]);
			switch (ch) {
				case 'H':
					printHelp=1;
				break;
				case 'Q':
					quiet=1;
				break;
				case 'M':
					if (GetStrParam(argc, argv, &i, P+2, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					
					P=st;
					model=-1;
					for (j=F84; j<numModels; j++) {
						if (strncmp(P, ModelNames[j], 3)==0)
							model=j;
					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}

	if (!quiet || printHelp)
 		PrintTitle();
	if (printHelp)
 		PrintUsage();
	SetModel(printHelp);
	if (printHelp)
		exit(0);
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (*P!='-') {
			if (treeFile) {
				fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
				PrintUsage();
				exit(0);
			}
			treeFile=1;
			strcpy(treeFileName, argv[i]);
		} else {
			while (*P) {
				*P=toupper(*P);
				P++;
			}
			P=argv[i];
			P++;
			ch=*P;
			P++;
			switch (ch) {
				case 'M': 	/* although this has already been read, we need to
							   use it up in this run through of the params.    */
					GetStrParam(argc, argv, &i, P, st, 3);
				break;
				case 'H':
				case 'Q':
				break;
				case 'R':
					if (GetIntParams(argc, argv, &j, P, 1, &numReps)) {
						fprintf(stderr, "Bad number of replicates: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'C':
					numCategories=3;
					for(j=0;j<numCategories;j++)
						codonCats[j]=j;
					codonRates=1;
					if (GetDoubleParams(argc, argv, &i, P, 3, siteRate)) {
						fprintf(stderr, "Bad Codon Category Rates: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					sumRates=siteRate[0]+siteRate[1]+siteRate[2];
					if (sumRates!=3.0) {
						siteRate[0]*=3.0/sumRates;
						siteRate[1]*=3.0/sumRates;
						siteRate[2]*=3.0/sumRates;
					}
				break;
				case 'G':
					if (GetIntParams(argc, argv, &i, P, 1, &numCategories) || numCategories<1 || numCategories>MAX_CATEGORIES) {
						fprintf(stderr, "Bad number of Gamma rate categories: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					gammaRates=1;
				break;
				case 'L':
					if (GetIntParams(argc, argv, &i, P, 1, &minWinSize) || minWinSize<1 ) {
						fprintf(stderr, "Bad minimum window size: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'A':
					if (gammaRates) {
						if (GetDoubleParams(argc, argv, &i, P, 1, &alpha) || alpha<0.0) {
							fprintf(stderr, "Bad alpha shape for Gamma rate heterogeneity: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else {
						fprintf(stderr, "Must specify the number of gamma rate categories before setting alpha: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'V':
					verbose=1;
					while (*P!='\0') {
						switch (*P) {
							case 'M': verboseMemory=1; break;
							case 'S': verboseSiteLik=1; break;
							default:					
								fprintf(stderr, "Unknown verbose mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				case 'D':
					debug=1;
					while (*P!='\0') {
						switch (*P) {
/*							case 'S': debugSplits=1; break;*/
							default:					
								fprintf(stderr, "Unknown debug mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				case 'Z':
					if (GetIntParams(argc, argv, &i, P, 2, &testRegions[numTestRegions*2])) {
						fprintf(stderr, "Bad test regions: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					numTestRegions++;
				break;
				default:
					if (ReadModelParam(argc, argv, &i)==-1) {
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}
	
	if (numCategories<2)
		gammaRates=0;

	if (FinishModelParam()<0) {
		PrintUsage();
		exit(0);
	}
}


void ReadFileParams()
{
	char ch, st[256];
	
	ch=fgetc(stdin);
	while (!feof(stdin) && isspace(ch)) 
		ch=fgetc(stdin);
	ungetc(ch, stdin);
	fgets(st, 255, stdin);
	if ( sscanf( st, " %d %d", &numTaxa, &numBases)!=2 ) {
		fprintf(stderr, "Unable to read parameters from standard input\n");
		exit(0);
	}
}


void ReadFile()
{
	int n, b, i;
	char ch;
		
	n=0;
	do {
		ch=fgetc(stdin);
		while ( !feof(stdin) && isspace(ch)) 
			ch=fgetc(stdin);
			
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n"); 
			exit(0);
		}
	
		i=0;
		while ( i<MAX_NAME_LEN && !feof(stdin) && !isspace(ch) ) {
			names[n][i]=ch;
			ch=fgetc(stdin);
			i++;
		}
		names[n][i]='\0';
		if (i==0) {
			fprintf(stderr, "Name missing for species %d\n", n+1);
			exit(0);
		}
		while (!feof(stdin) && isspace(ch))
			ch=fgetc(stdin);
		
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		b=0;
		while ( !feof(stdin) && b<numBases) {
			if ( !isspace(ch) ) {
				sequences[n][b]=ch;
				b++;
			}
			ch=toupper(fgetc(stdin));
		}
		
		if ( b<numBases ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		/*fprintf(stderr, "%d: %s, bases read: %d\n", n+1, names[n], b);*/
		n++;
		
		if ( n<numTaxa && feof(stdin) ) {
			fprintf(stderr, "Too few sequences in input file\n");
			exit(0);
		}
	} while ( n<numTaxa );
}


int OpenTreeFile()
{
	char st[256];
	int n;
		
	if (treeFile) {
		if ( (tree_fv=fopen(treeFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening tree file: '%s'\n", treeFileName);
			exit(0);
		}
		n=0;
		while (!feof(tree_fv)) {
			if (fgetc(tree_fv)==';')
				n++;
		}
		rewind(tree_fv);
	} else {
		tree_fv=NULL;
		fgets(st, 255, stdin);
		if ( sscanf(st, " %d ", &n)!=1 ) {
			fprintf(stderr, "Tree is missing from end of sequence file\n");
			exit(0);
		}
	}
	
	return n;
}

TTree *ReadNextTree(TTree *tree)
{		
	if (treeFile) {
		ReadTree(tree_fv, tree);
	} else {
		ReadTree(stdin, tree);
	}
	
	return tree;
}

void CloseTreeFile()
{
	if (treeFile)
		fclose(tree_fv);
}

void PrintVerbose(TTree *tree)
{
	int i;
	
	fprintf(stderr, "Number of null simulations = %d\n", numReps);
	fprintf(stderr, "%d taxa, %d bases\n", numTaxa, numBases);
	fprintf(stderr, "	%d distinct patterns\n\n", numPatterns);
	if (numCategories>1) {
		fprintf(stderr, "Rate heterogeneity:\n");
		if (gammaRates) {
			fprintf(stderr, "  Discrete Gamma rate heterogeneity with %d categories\n", numCategories);
			fprintf(stderr, "  shape parameter, alpha = %f\n", alpha);
		} else {
			fprintf(stderr, "  Codon specific rates supplied by user\n");
		}
		fprintf(stderr, "    relative rates (mean set to 1.0) =");
		for (i=0; i<numCategories; i++)
			fprintf(stderr, " %f", siteRate[i]);
		fputc('\n', stderr);
	} else
		fprintf(stderr, "Rate homogeneity assumed");
	fprintf(stderr, ".\n");
	fprintf(stderr, "Model=%s\n", ModelNames[model]);
}
	
	
void PrintSiteLik(FILE *fv)
{
	int j;

	fprintf(fv, "1");
	if (codonRates) {
		for(j=0;j<numBases;j++){
			fprintf(fv, "\t%f", log(patternLik[codonCats[j % 3]][patternIndex[j]]));
			if ((j+1)%6==0) {
				fprintf(fv, "\n%d", j+2);
			}
		}
	} else {
		for(j=0;j<numBases;j++){
			fprintf(fv, "\t%f", log(patternLik[0][patternIndex[j]]));
			if ((j+1)%6==0) {
				fprintf(fv, "\n%d", j+2);
			}
		}
	}
	fprintf(fv, "\n\n");
}


int main(int argc, char **argv)
{
	TTree *tree;
	clock_t coreStart, totalStart;
	double coreSecs, totalSecs;
#if __profile__
	OSErr  theErr;
#endif
	

#ifdef __MWERKS__
	argc=ccommand(&argv);
#endif
	
	SetSeed(time(NULL));

	totalMem=0;
 	
	totalStart = clock();
	
	estimateGlobals=0;
	numModelCats=1;

	ReadParams(argc, argv);

	ReadFileParams();
	
	AllocateMemory();
		
	ReadFile();
	
	if (gammaRates)
		DiscreteGamma(freqRate, siteRate, alpha, alpha, numCategories, 0);

	CreatePatterns();

	tree=CreateTree(0);
		
	numTrees=OpenTreeFile();
	ReadNextTree(tree);
	if (numTrees!=1){
		fprintf(stderr, "Only one tree can be tested at a time\n");
		exit(0);
	}
	numSets=numTrees;			
	
	if (!quiet && verbose)
 		PrintVerbose(tree);

	AllocateTreeLikelihoods(tree);
	SetTipLikelihoods(tree);
	SetupLikelihood(tree);
	 
	coreStart = clock();

	#if __profile__
		theErr = ProfilerInit( 0, 4, 40, 40 );
		if ( theErr != noErr ) {
			fprintf(stderr, "Profiling failed to start\n");
			exit(0);
		}       
	#endif

	tree->lnL=CalcLik();
	
	if (!quiet)
		fprintf(stderr, "Tree -ln Likelihood = %f\n", tree->lnL);

	if (verboseSiteLik)
		PrintSiteLik(stdout);

	Anomaly(tree);
	
	#if __profile__
		ProfilerDump("\pspatula.profile");
		ProfilerTerm();
	#endif

	coreSecs = (double)(clock() - coreStart) / CLOCKS_PER_SEC;		
	CloseTreeFile();
	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
	if (!quiet) {
		fprintf(stderr, "Time taken: %G seconds\n", totalSecs);
		if (verboseMemory)
			fprintf(stderr, "Total memory used: %ld\n", totalMem);
	}
	return 0;
}



