/*  (c) Copyright 1996, Andrew Rambaut & Nick Grassly  */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "lik.h"
#include "tree.h"


static TNode *nextFreeNode;
static int nextBranch, nextSplit;


/* prototypes */

char ReadToNextChar(FILE *fv);
void ReadUntil(FILE *fv, char stopChar, char *what);
TNode *ReadTip(FILE *fv, char ch, TTree *tree);
TNode *ReadNode(FILE *fv, TNode *node, TTree *tree);
void ConnectBranch(TNode *destNode, TNode *srcNode);
TNode *ReadBranch(FILE *fv, TTree *tree);

void WriteUnrootedNode(FILE *fv, TNode *node);


/* functions */

TTree *CreateTree(short isRooted)
{
	TTree *tree;
	TNode *P;
	int i;
	
	tree=(TTree *)AllocMem(sizeof(TTree), "tree", "CreateTree", 1);

	tree->isRooted=isRooted;
	tree->numTips=numTaxa;
	tree->numNodes=numTaxa-2;
	tree->totalNodes=numTaxa+(3*tree->numNodes);

	tree->nodes=(TNode *)AllocMem(sizeof(TNode)*tree->totalNodes, "tree->nodes", "CreateTree", 1);

	P=tree->nodes;
	for (i=0; i<tree->numTips; i++) {
		P->branch=NULL;
		P->next=NULL;
		
		P->Likelihood=NULL;	
		P->tipNo=i;
		P->nodeNo=-1;
		P->sequence=NULL;
		P++;
	}	
	
	tree->start=tree->nodes;
	
	for (i=0; i<tree->numNodes; i++) {
		
		P->Likelihood=NULL;
		P->branch=NULL;
		P->tipNo=-1;
		P->nodeNo=i;
		P->next=P+1;
		P->sequence=NULL;
		
		(P+1)->branch=NULL;
		(P+1)->Likelihood=NULL;
		(P+1)->tipNo=-1;
		(P+1)->nodeNo=i;
		(P+1)->next=P+2;
		(P+1)->sequence=NULL;	
		
		(P+2)->branch=NULL;
		(P+2)->Likelihood=NULL;
		(P+2)->tipNo=-1;
		(P+2)->nodeNo=i;
		(P+2)->next=P;
		(P+2)->sequence=NULL;
		
		P+=3;
	}
	
	tree->branches=(TNode **)AllocMem(sizeof(TNode *)*(numBranches+1), "tree->branches", "CreateTree", 1);
	
	return tree;
}


void DisposeTree(TTree *tree)
{
	int i;
	
	if (tree!=NULL) {
		for (i=0; i<tree->totalNodes; i++)
			free(tree->nodes[i].Likelihood);
		
		if (tree->nodes!=NULL)
			free(tree->nodes);
		
		if (tree->branches!=NULL)
			free(tree->branches);
			
		free(tree);
	}
}


char ReadToNextChar(FILE *fv)
{
	char ch;
	
	ch=fgetc(fv);
	while (!feof(fv) && isspace(ch)) 
		ch=fgetc(fv);

	return ch;
}


void ReadUntil(FILE *fv, char stopChar, char *what)
{
	char ch;
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=stopChar) 
		ch=fgetc(fv);

	if (feof(fv)) {
		fprintf(stderr, "%s missing in tree file\n", what);
		exit(0);
	}
}


TNode *ReadTip(FILE *fv, char ch, TTree *tree)
{
	int i;
	char st[MAX_NAME_LEN+1];
	TNode *node;
	
	i=0;
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')' && i<MAX_NAME_LEN) { 
		st[i]=ch;
		i++;
		ch=fgetc(fv);
	}
	st[i]='\0';

	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')') 
		ch=fgetc(fv);
		
	if (feof(fv)) {
		fprintf(stderr, "Unexpected end of file for tree\n");
		exit(0);
	}

	ungetc(ch, fv);
	
	if ((i=FindTipName(st))<0) {
		fprintf(stderr, "Unknown taxa lable in tree: %s\n", st);
		exit(0);
	}
	node=(tree->nodes)+i;
	
	return node;
}


TNode *ReadNode(FILE *fv, TNode *node, TTree *tree)
{
	char ch;
	TNode *node2;
	
	node2=ReadBranch(fv, tree);
	ConnectBranch(node->next, node2);
	ReadUntil(fv, ',', "Comma");
	
	node2=ReadBranch(fv, tree);
	ConnectBranch(node->next->next, node2);
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=')') 
		ch=fgetc(fv);

	ch=fgetc(fv);
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')') 
		ch=fgetc(fv);
		
	if (feof(fv)) {
		fprintf(stderr, "Unexpected end of file for tree\n");
		exit(0);
	}

	ungetc(ch, fv);
	
	return node;
}


void ConnectBranch(TNode *destNode, TNode *srcNode)
{
	srcNode->branch=destNode;
	destNode->branchNo=srcNode->branchNo;
	destNode->branchLength=srcNode->branchLength;
	destNode->branch=srcNode;
}	


TNode *ReadBranch(FILE *fv, TTree *tree)
{
	char ch;
	double dist;
	TNode *node;
	
	ch=ReadToNextChar(fv);
	if (ch=='(') {	/* is a node */
		node=nextFreeNode;
		nextFreeNode+=3;
		node=ReadNode(fv, node, tree);
	} else {		/* is a tip */
		node=ReadTip(fv, ch, tree);
	}
	
	ch=fgetc(fv);
	if (ch==':') {
		if (fscanf(fv, "%lf", &dist)!=1) {
			fprintf(stderr, "Bad branch length in tree.\n");
			exit(0);
		}
	} else {
		dist=0.0;
		ungetc(ch, fv);
	}
	
/* we must allow an extra branch length in branchValue in case */
/* the tree is rooted. In this case it will be summed with the */
/* branch on the other side of the root */
	node->branchNo=nextBranch;
	branchValue[nextBranch]=dist;
	node->branchLength=branchValue+nextBranch;
	tree->branches[nextBranch]=node;
		
	nextBranch++;
	
	return node;
}	


void ReadTree(FILE *fv, TTree *tree)
{
	char ch;
	TNode *node1, *node2, *node3;
	
	nextFreeNode=tree->nodes+tree->numTips;
	nextBranch=0;
	
	tree->isRooted=0;
	
	ReadUntil(fv, '(', "tree");
	node1=ReadBranch(fv, tree);
	ReadUntil(fv, ',', "Comma");
	node2=ReadBranch(fv, tree);
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=',' && ch!=')') 
		ch=fgetc(fv);

	if (ch==',') {		/* the tree is unrooted */
		node3=ReadBranch(fv, tree);

		ConnectBranch(nextFreeNode, node1);
		ConnectBranch(nextFreeNode->next, node2);
		ConnectBranch(nextFreeNode->next->next, node3);
	} else if (ch==')') {  			/* the tree is rooted - convert it to unrooted */		
		(*node1->branchLength)+=*node2->branchLength;
		ConnectBranch(node2, node1);
	} else {
		fprintf(stderr, "Closing bracket missing in tree file.\n");
		exit(0);
	}
	
	ReadUntil(fv, ';', "Semi-colon");

	tree->start=tree->nodes[0].branch;
}


void WriteUnrootedNode(FILE *fv, TNode *node)
{
	if (node->tipNo<0) {
		fprintf(fv, "(");
		WriteUnrootedNode(fv, node->next->branch);
		fprintf(fv, ":%.8f,", *node->next->branchLength);
		WriteUnrootedNode(fv, node->next->next->branch);
		fprintf(fv, ":%.8f)", *node->next->next->branchLength);
	} else {
		fprintf(fv, "%s", names[node->tipNo]);
	}
}


void WriteTree(FILE *fv, TTree *tree, int printRates)
{
	fprintf(fv, "(");
	WriteUnrootedNode(fv, tree->start->next->branch);
	fprintf(fv, ":%.8f,", *tree->start->next->branchLength);
	WriteUnrootedNode(fv, tree->start->next->next->branch);
	fprintf(fv, ":%.8f,", *tree->start->next->next->branchLength);
	WriteUnrootedNode(fv, tree->start->branch);
	fprintf(fv, ":%.8f);\n", *tree->start->branchLength);
}
